from datetime import datetime, timedelta
from dateutil import tz

import logging

log = logging.getLogger('s3_backups')

"""
The default schedule does the following ...

- Keeps all archives for 7 days
- Keeps midnight backups for every other day for 30 days
- Keeps the first day of the month forever
"""

ONE_WEEK = timedelta(days=7)
ONE_MONTH = timedelta(days=30)
utc = tz.tzutc()


class OddEven:

    current = 'even'

    def toggle(self):
        if self.current == 'even':
            self.current = 'odd'
        else:
            self.current = 'even'
        return self.current

oe = OddEven()


def keep_file(obj):
    utcnow = datetime.utcnow()
    utcnow = utcnow.replace(tzinfo=utc)
    timediff = utcnow - obj.utc_last_modified
    odd_even = oe.toggle()

    if timediff <= ONE_WEEK:
        log.info("%s - Keeping key \"%s\" because it's less than a week old." % (timediff, obj.key))
        return True
    elif timediff > ONE_WEEK and timediff < ONE_MONTH:
        if obj.local_last_modified.hour != 0:
            log.info("%s - Removing key \"%s\" because it's not a midnight backup and it's older than one week but less than a month" % (timediff, obj.key))
            return False
        elif odd_even == 'even':
            log.info("%s - Removing key \"%s\" because it's older than one week but less than a month and not an even day." % (timediff, obj.key))
            return False
        else:
            log.info("%s - Keeping key \"%s\" because it's older than one week but less than a month and it's an even day." % (timediff, obj.key))
            return True
    elif timediff > ONE_MONTH:
        if obj.local_last_modified.day == 1:
            log.info("%s - Keeping key \"%s\" because it's older than a month and also the first day of the month." % (timediff, obj.key))
            return True
        else:
            log.info("%s - Removing key \"%s\" because it's older than a month and not the first day of the month." % (timediff, obj.key))
            return False



from datetime import datetime

import re


class Archive(object):
    """
    Archives all backups on S3 using the following schedule:

    - Keep all backups for 7 days
    - Keep midnight backups for every other day for 30 days
    - Keep 1st day of the month forever
    """

    def __init__(self, conn, SERVICE_NAME, BUCKET_NAME, FILE_KEY):
        self.conn = conn
        self.SERVICE_NAME = SERVICE_NAME
        self.BUCKET_NAME = BUCKET_NAME
        self.FILE_KEY = FILE_KEY

    def schedule(self, conn, schedule_module='schedule'):
        schedule = __import__(schedule_module)
        bucket = self.conn[self.SERVICE_NAME].get_bucket(self.BUCKET_NAME)

        file_key = self.FILE_KEY if self.FILE_KEY.endswith('/') else self.FILE_KEY + '/'

        for obj in bucket.objects.filter(Prefix=file_key):
            if not obj.key.endswith("/"):

                obj = self.add_datetimes_to_key(obj)

                # create a new key that puts the archive in a year/mon`th sub
                # directory if it's not in a year/month sub directory already
                name_parts = obj.key.split('/')
                month = name_parts[-2]
                year = name_parts[-3]
                new_key_name = obj.key
                if not re.match(r'[\d]{4}', year) and not re.match(r'[\d]{2}', month):
                    name_parts.insert(len(name_parts) - 1, "%d" % obj.local_last_modified.year)
                    name_parts.insert(len(name_parts) - 1, "%02d" % obj.local_last_modified.month)
                    new_key_name = "/".join(name_parts)

                # either keep the file or delete it
                keep_file = schedule.keep_file(obj)
                if keep_file and obj.key != new_key_name:
                    self.conn.Object(self.BUCKET_NAME, new_key_name, metadata=key.meta, preserve_acl=True).copy_from(CopySource=obj.key)
                    bucket.delete_key(obj.key)
                elif not keep_file:
                    bucket.delete_key(obj.key)

    def add_datetimes_to_key(self, obj):
        """
        Convert the last_modified GMT datetime string to a datetime object and
        create utc and local datetime objects.
        """

        utc = tz.tzutc()
        gmt = tz.gettz('GMT')
        local_tz = tz.tzlocal()

        obj.last_modified = datetime.strptime(obj.last_modified, "%Y-%m-%dT%H:%M:%S.%fZ")
        obj.last_modified = obj.last_modified.replace(tzinfo=gmt)
        obj.utc_last_modified = obj.last_modified.astimezone(utc)
        obj.local_last_modified = obj.last_modified.astimezone(local_tz)

        return obj



import os
import boto3


class s3Connect:
    """
    Class for lazy connections.
    That is make connections
      a) only when they are needed and then cache them
      b) not while loading the module
    """
    aws = None
    do = None

    def __init__(self, ACCESS_KEY, SECRET, REGION):
        self.ACCESS_KEY = ACCESS_KEY
        self.SECRET = SECRET
        self.REGION = REGION

    def make_aws_conn(self):
        AWS_BASE_URL = 'https://s3.amazonaws.com'
        return boto3.resource('s3',
                              region_name=self.REGION,
                              endpoint_url=AWS_BASE_URL,
                              aws_access_key_id=self.ACCESS_KEY,
                              aws_secret_access_key=self.SECRET)

    def make_do_conn(self):
        DO_BASE_URL = 'https://%s.%s' % (self.REGION, 'digitaloceanspaces.com')
        return boto3.resource('s3',
                              region_name=self.REGION,
                              endpoint_url=DO_BASE_URL,
                              aws_access_key_id=self.ACCESS_KEY,
                              aws_secret_access_key=self.SECRET)

    def __getitem__(self, name):
        conn = None
        if name == 'amazon':
            if self.aws is None:
                self.aws = self.make_aws_conn()
            conn = self.aws
        elif name == 'digitalocean':
            if self.do is None:
                self.do = self.make_do_conn()
            conn = self.do
        return conn

    def upload_file_to_cloud(self, service, bucket, logger,
                             media_location, file_key, DELETE_DUMP):
        """
        Uploads file to Cloud (Amazon or DigitalOcean) S3 services
        Arguments:
            media_location -- File location in local storage
            file_key -- file key, the directory path where the file is stored
            service -- 'amazon' for Amazon or 'digitalocean'
            bucket -- bucket name
        """
        logger.info('Uploading file key {} to Cloud ...'.format(file_key))
        buck = self.__getitem__(service).Bucket(bucket)
        buck.put_object(
            Key=file_key,
            Body=open(media_location, 'rb'),
            ACL='public-read')
        logger.info('Uploaded file key {}.'.format(file_key))
        if DELETE_DUMP:
            os.remove(media_location)
            logger.info('''Removed the dump from
                        local directory ({}).'''.format(media_location))



#!/usr/bin/env python

from datetime import datetime
from connect import s3Connect
from archive import Archive

import os
import argparse
import subprocess

import utils

import logging
import tarfile
import tempfile

logger = logging.getLogger('s3_backups')


def create_db_dump(command, filename):
    """
    Creates dump file using pg_dump
    Returns:
        1: Path of dump file from temp storage
    """
    if not os.path.exists(DUMP_BASE_DIR):
        os.mkdir(DUMP_BASE_DIR)

    logger.info("Preparing " + filename + ".dump from the database dump ...")
    with tempfile.NamedTemporaryFile() as temp1:
        ps = subprocess.Popen(
            command,
            stdout=temp1, shell=True, universal_newlines=True
        )
        ps.wait()
        temp1.flush()

        tar = tarfile.open(DUMP_BASE_DIR + filename + ".tar.gz", "w|gz")
        tar.add(temp1.name, filename + ".sql")
        tar.close()
    logger.info("Created tar file " + filename + ".tar.gz")
    return '{}{}.tar.gz'.format(DUMP_BASE_DIR, filename)


def backup():
    """Creates the buckup and uploads"""
    now = datetime.now()
    filename = ARCHIVE_NAME + now.strftime('_%Y%m%d_%H%M%S')
    if DB_NAME:
        command = [
            POSTGRES_DUMP_CMD,
            '-Fc', DB_NAME,
            '-f', DUMP_BASE_DIR + filename + ".sql"
        ]
    else:
        command = [
            POSTGRES_DUMP_CMD + 'all',
            '-f', DUMP_BASE_DIR + filename + ".sql"
        ]

    file_location = create_db_dump(command, filename)
    file_key = utils.get_file_key(FILE_KEY, filename, ARCHIVE)

    conn.upload_file_to_cloud(
        media_location=file_location,
        file_key=file_key,
        service=SERVICE_NAME,
        bucket=BUCKET_NAME,
        DELETE_DUMP=DELETE_DUMP
    )
    logger.info('Sucessfully uploaded the Dump.')


if __name__ == '__main__':

    parser = argparse.ArgumentParser(description='''Creates postgres DB
                                     dump and stores it to S3 using
                                     `pg_dump`.''')
    parser = utils.init_arguments(parser)

    parser.add_argument('--ARCHIVE_NAME',
                        default='all_postgres_db',
                        help='The base name for the archive')

    args = parser.parse_args()

    SERVICE_NAME = args.SERVICE_NAME
    ACCESS_KEY = args.ACCESS_KEY
    SECRET = args.SECRET
    REGION = args.REGION
    BUCKET_NAME = args.BUCKET_NAME
    FILE_KEY = args.FILE_KEY

    DB_NAME = args.DB_NAME
    DELETE_DUMP = args.DELETE_DUMP
    DUMP_BASE_DIR = args.DUMP_BASE_DIR
    DUMP_FILE_PATH = args.DUMP_FILE_PATH
    ARCHIVE_NAME = args.ARCHIVE_NAME

    POSTGRES_DUMP_CMD = args.POSTGRES_DUMP_CMD
    conn = s3Connect()

    if args.archive:
        archive = Archive(conn, SERVICE_NAME=SERVICE_NAME, BUCKET_NAME=BUCKET_NAME, FILE_KEY=FILE_KEY)
        archive.schedule()

    if args.verbose:
        utils.init_logger()

    if args.backup:
        backup()



#!/usr/bin/env python

from datetime import datetime
from connect import s3Connect

import os
import argparse
import subprocess

import utils

import logging
import tarfile

logger = logging.getLogger('s3_backups')


def create_redis_dump(filename):
    """
    Creates dump file using pg_dump
    Returns:
        1: Path of dump file from temp storage
    """
    if not os.path.exists(DUMP_BASE_DIR):
        os.mkdir(DUMP_BASE_DIR)

    logger.info("Preparing " + filename + ".dump from the database dump ...")
    ps = subprocess.Popen(
            REDIS_SAVE_CMD, shell=True, stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )
    ps.wait()

    tar = tarfile.open(DUMP_BASE_DIR + filename + ".tar.gz", "w|gz")
    tar.add(DUMP_RDB_PATH, filename + ".rdb")
    tar.close()
    logger.info("Created tar file " + filename + ".tar.gz")
    return '{}{}.tar.gz'.format(DUMP_BASE_DIR, filename)


def backup():
    """Creates the buckup and uploads"""
    now = datetime.now()
    filename = ARCHIVE_NAME + now.strftime('_%Y%m%d_%H%M%S')

    file_location = create_redis_dump(filename)
    file_key = utils.get_file_key(FILE_KEY, filename, ARCHIVE)

    conn.upload_file_to_cloud(
        media_location=file_location,
        file_key=file_key,
        service=SERVICE_NAME,
        bucket=BUCKET_NAME
    )

    if DELETE_DUMP:
        os.removedirs(DUMP_BASE_DIR)

    logger.info('Sucessfully uploaded the Dump.')


if __name__ == '__main__':

    parser = argparse.ArgumentParser(
                description='''Creates redis DB dump and stores
                it to S3 using `redis-cli save`.''')
    parser = utils.init_arguments(parser)
    parser.add_argument('--ARCHIVE_NAME',
                        default='all_redis_db',
                        help='The base name for the archive')

    args = parser.parse_args()

    SERVICE_NAME = args.SERVICE_NAME
    ACCESS_KEY = args.ACCESS_KEY
    SECRET = args.SECRET
    REGION = args.REGION
    BUCKET_NAME = args.BUCKET_NAME
    FILE_KEY = args.FILE_KEY

    DELETE_DUMP = args.DELETE_DUMP
    DUMP_BASE_DIR = args.DUMP_BASE_DIR

    DUMP_RDB_PATH = args.DUMP_RDB_PATH
    REDIS_SAVE_CMD = args.REDIS_SAVE_CMD
    ARCHIVE_NAME = args.ARCHIVE_NAME

    conn = s3Connect(ACCESS_KEY, SECRET, REGION)

    if args.archive:
        archive = Archive(conn, SERVICE_NAME=SERVICE_NAME, BUCKET_NAME=BUCKET_NAME, FILE_KEY=FILE_KEY)
        archive.schedule()

    if args.verbose:
        utils.init_logger()

    if args.backup:
        backup()



from setuptools import setup

setup(
    name='s3_dumps',
    packages=['s3_dumps'],
    version=__import__('s3_dumps').__version__,
    description='A package for backup DB and store in s3',
    author='Reck31',
    author_email='rakesh.gunduka@gmail.com',
    url='https://github.com/rakeshgunduka/s3_dumps',
    include_package_data=True,
    license=open('LICENSE.txt').read(),
    install_requires=[
        "boto3"
    ],
    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
    ],
    entry_points={
        'console_scripts': [
            'postgres_to_s3=s3_dumps.postgres_to_s3:main',
            'redis_to_s3=s3_dumps.redis_to_s3:main'
        ],
    },
    scripts=[
        's3_dumps/postgres_to_s3.py',
        's3_dumps/redis_to_s3.py',
        's3_dumps/connect.py',
        's3_dumps/utils.py'
    ],
    keywords = ['s3', 'backup', 'postgres', 'redis', 'digitaloceanspaces', 'aws'],
    zip_safe=False
)



from datetime import datetime

import os
import logging


def init_arguments(parser):
    SERVICE_NAME = os.environ.get('SERVICE_NAME')
    ACCESS_KEY = os.environ.get('ACCESS_KEY')
    SECRET = os.environ.get('SECRET')

    parser.add_argument('--SERVICE_NAME',
                        required=SERVICE_NAME is None,
                        default=SERVICE_NAME,
                        help='''Supported Amazon/ DigitalOcean Services Eg.
                        amazon or digitalocean ( field is required if not
                        defined in SERVICE_NAME environment variable )''')

    parser.add_argument('--ACCESS_KEY',
                        required=ACCESS_KEY is None,
                        default=ACCESS_KEY,
                        help='''Access key ( field is required if not defined
                        in ACCESS_KEY environment variable )''')

    parser.add_argument('--SECRET',
                        required=SECRET is None,
                        default=SECRET,
                        help='''Secret key (required if not defined in SECRET
                        environment variable)''')

    parser.add_argument('--BUCKET_NAME',
                        required=True,
                        help='Bucket name')

    parser.add_argument('--FILE_KEY',
                        required=True,
                        help='''The directory path where you want to put dump
                        (i.e. backups/postgres/server_name)''')

    parser.add_argument('--REGION',
                        default='',
                        help='s3 region Eg. us-east-1 (default is set to "")')

    parser.add_argument('--DB_NAME',
                        required=False,
                        help='''The database you want to backup
                        (Eg. my_db_name). If DB_NAME not provided then
                        it dumps all databases.''')

    parser.add_argument('--POSTGRES_DUMP_CMD',
                        default='pg_dump',
                        help='''Path to pg_dumpall (default: pg_dump).
                        You may change according to system
                        Eg. /usr/bin/pg_dump''')

    parser.add_argument('--DUMP_BASE_DIR',
                        default='',
                        help='Path to dumps directory (default: "")')

    parser.add_argument('--DELETE_DUMP',
                        default=True,
                        help='''Boolean value if True deletes the dump file
                        from dumps base directory''')

    parser.add_argument('--DUMP_RDB_PATH',
                        default='/var/lib/redis/dump.rdb',
                        help='''The path to the Redis dump.rdb file
                            (default: /var/lib/redis/dump.rdb)''')

    parser.add_argument('--REDIS_SAVE_CMD',
                        default='/usr/bin/redis-cli SAVE',
                        help='''Command to save the Redis DB to disk
                        (default: /usr/bin/redis-cli SAVE)''')

    parser.add_argument('-v', '--verbose',
                        default=True,
                        action='store_true',
                        help='Verbose')

    parser.add_argument('--backup',
                        action='store_true',
                        help='Backup up Postgres to S3')

    parser.add_argument('--archive',
                        action='store_true',
                        help='Backup up Postgres to S3')

    return parser


def init_logger(logger):
    logger.setLevel(logging.INFO)
    ch = logging.StreamHandler()
    formatter = logging.Formatter('''
        %(asctime)s - %(name)s - %(levelname)s - %(message)s''')
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    return logger


def get_file_key(file_key, filename, ARCHIVE):
    file_key_suffix = file_key if not file_key.endswith('/') else file_key[:-1]
    if ARCHIVE:
        file_key_suffix += datetime.now().strftime('/%Y/%m/%d')
    return r'%s/%s' % (file_key_suffix, filename + '.tar.gz')
